//
//  JSCoreExecutor.m
//  Runner
//
//  Created by 苏玮 on 2019/12/9.
//  Copyright © 2019 The Chromium Authors. All rights reserved.
//

#import <Foundation/Foundation.h>
#import <JavaScriptCore/JavaScriptCore.h>
#import "JSCoreExecutor.h"
#import <pthread.h>
#import <UIKit/UIDevice.h>

@interface JSCoreContext : NSObject

@property (nonatomic, strong, readonly) JSContext *context;

- (instancetype)initWithJSContext:(JSContext *)context
                         onThread:(NSThread *)javaScriptThread;

@end


@implementation JSCoreContext
{

    NSThread *_javaScriptThread;
}

- (instancetype)initWithJSContext:(JSContext *)context
                         onThread:(NSThread *)javaScriptThread
{
    if ((self = [super init])) {
        _context = context;
        _context.name = @"DPOFlutterJSContext";
        _javaScriptThread = javaScriptThread;
    }
    return self;
}


- (BOOL)isValid
{
    return _context != nil;
}

- (void)invalidate
{
    if (self.isValid) {
        _context = nil;
        _javaScriptThread = nil;
        
        CFRunLoopStop([[NSRunLoop currentRunLoop] getCFRunLoop]);
    }
}

@end

@implementation JSCoreExecutor
{
    JSCoreContext *_jsContext;
    NSThread *_javaScriptThread;
}

- (instancetype)init
{
    if (self = [super init])
    {
        [self setup];
    }
    return self;
}

- (void)dealloc
{
    [_jsContext performSelector:@selector(invalidate)
                     onThread:_javaScriptThread
                   withObject:nil
                waitUntilDone:NO];
    _jsContext = nil;
    // TODO :emit log
    // DPOFlutterLog(@"dealloc ");
    
}


- (JSContext *)jsContext
{
    return _jsContext.context;
}

- (void)setup
{
    NSThread *javaScriptThread = [[NSThread alloc] initWithTarget:[JSCoreExecutor class]
                                                         selector:@selector(runRunLoopThread)
                                                           object:nil];
    javaScriptThread.name = @"__DPOFlutterJSThread__";
    javaScriptThread.qualityOfService = NSOperationQualityOfServiceUserInteractive;
    [javaScriptThread start];
    
    _javaScriptThread = javaScriptThread;
    
    __weak JSCoreExecutor *weakSelf = self;
    [self executeBlockOnJSThread:^{
        
        JSCoreExecutor *strongSelf = weakSelf;
        
        if (strongSelf == nil)
        {
            return ;
        }
        JSContext *jsContext = [[JSContext alloc] init];
        strongSelf->_jsContext = [[JSCoreContext alloc] initWithJSContext:jsContext onThread:strongSelf->_javaScriptThread];
        strongSelf->_isValid = YES;
       
    }];
   
}


+ (void)runRunLoopThread
{
    @autoreleasepool {
      
        pthread_setname_np([NSThread currentThread].name.UTF8String);
        
        CFRunLoopSourceContext noSpinCtx = {0, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
        CFRunLoopSourceRef noSpinSource = CFRunLoopSourceCreate(NULL, 0, &noSpinCtx);
        CFRunLoopAddSource(CFRunLoopGetCurrent(), noSpinSource, kCFRunLoopDefaultMode);
        CFRelease(noSpinSource);
 
        while (kCFRunLoopRunStopped != CFRunLoopRunInMode(kCFRunLoopDefaultMode, ((NSDate *)[NSDate distantFuture]).timeIntervalSinceReferenceDate, NO)) {
            
        }
    }
}


- (JSCoreContext *)context
{
    if (!self.isValid)
    {
        return nil;
    }

    return _jsContext;
}



- (void)executeBlockOnJSThread:(dispatch_block_t)block
{
    if ([NSThread currentThread] != _javaScriptThread) {
        [self performSelector:@selector(executeBlockOnJSThread:)
                     onThread:_javaScriptThread withObject:block waitUntilDone:NO];
    } else {
        block();
    }
}

- (void)executeAsyncBlockOnJSThread:(dispatch_block_t)block
{
    [self performSelector:@selector(executeBlockOnJSThread:)
                 onThread:_javaScriptThread
               withObject:block
            waitUntilDone:NO];
}

- (void)executeJSCBlockOnJSThread:(JSCDispatchBlock)block
{
    if ([NSThread currentThread] != _javaScriptThread) {
        [self performSelector:@selector(executeJSCBlockOnJSThread:)
                     onThread:_javaScriptThread withObject:block waitUntilDone:NO];
    } else {
        block(self);
    }
}


- (JSValue *)executeJS:(NSString *)script
{
    JSValue *value = nil;
    value = [_jsContext.context evaluateScript:script];
    return value;
}

- (JSValue *)executeScript:(NSString *)script  sourceURL:(NSURL *)sourceURL
{
    JSValue *value = nil;
    if ([_jsContext.context respondsToSelector:@selector(evaluateScript:withSourceURL:)])
    {
        value = [_jsContext.context evaluateScript:script withSourceURL:sourceURL];
    } else
    {
        value = [_jsContext.context evaluateScript:script];
    }
    
    return value;
}

- (void)executeScriptAsync:(NSString *)script
                       sourceURL:(NSURL *)sourceURL
                      onComplete:(JSCCompleteBlock)onComplete
{

    NSError *error = nil;
    
    [self executeBlockOnJSThread:^{

        [self executeScript:script sourceURL:sourceURL];
        if (onComplete) {
            onComplete(error);
        }
    }];
}


- (void )invokeJSValue:(JSValue *)jsValue mothod:(NSString *)method args:(NSArray *)args callback:(JSCValueCallback )callback
{
    [self executeBlockOnJSThread:^{
        
        JSValue * result =   [jsValue invokeMethod:method withArguments:args];
        
        if (callback) {
            callback(result,nil);
        }
    }];
}

- (void)invokeMothod:(NSString *)method args:(NSArray *)args callback:(JSCValueCallback )callback
{
    [self executeBlockOnJSThread:^{
        
        JSValue * result =   [self.jsContext.globalObject invokeMethod:method withArguments:args];
        
        if (callback) {
            callback(result,nil);
        }
    }];
}

@end
