import 'dart:async';
import 'dart:convert';

import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:hybrid/ui/ui_factory.dart';
import 'package:hybrid/util/base64.dart';
import 'package:hybrid/util/color_util.dart';
import 'abstracts.dart';

class DpoMiniAppPage extends StatefulWidget {
  final Map<String, dynamic> args;
  final Map<String, dynamic> pages;
  final Map<String, HybridMessageHandler> handlers;
  final MethodChannel methodChannel;
  final BasicMessageChannel<String> basicChannel;
  final String indexPage;
  DpoMiniAppPage(this.args, this.pages, this.handlers, this.methodChannel,
      this.basicChannel, this.indexPage);

  @override
  DpoMiniAppPageState createState() => DpoMiniAppPageState(
      args, pages, handlers, methodChannel, basicChannel, indexPage);
}

class DpoMiniAppPageState extends State<DpoMiniAppPage>
    with HybridMessageHandler {
  Map<String, dynamic> _args = Map();
  Map<String, String> _pages = Map();
  Map<String, HybridMessageHandler> _handlers = Map();
  String _indexPage = "";
  Map<String, dynamic> _data;
  MethodChannel _methodChannel;
  BasicMessageChannel<String> _basicChannel;

  Color _appBarColor;
  Color _backgroundColor;
  bool _enablePullDownRefresh;

  String _pageCode = "";
  String _pageId = "";
  String _title = "";
  Widget _tree;

  GlobalKey<RefreshIndicatorState> _refreshIndicatorKey;
  Completer<bool> _completer;

  UIFactory _factory;

  DpoMiniAppPageState(this._args, this._pages, this._handlers,
      this._methodChannel, this._basicChannel, this._indexPage) {
    if (_args.containsKey("pageCode")) {
      _pageCode = _args['pageCode'];
      _args = _args['args'];
    } else {
      _pageCode = _indexPage;
    }
    _appBarColor = Colors.blue;
    _backgroundColor = Colors.grey[200];
    _enablePullDownRefresh = false;
    _pageId = _pageCode + this.hashCode.toString();
    _factory = UIFactory(_pageId, _methodChannel);
    _handlers.putIfAbsent(_pageId, () => this);
    _initBasicChannel();
  }

  _initBasicChannel() async {
    _basicChannel.setMessageHandler((String message) {
      // print('Flutter Received: $message');
      var jsonObj = jsonDecode(message);
      var pageId = jsonObj['pageId'];
      HybridMessageHandler handler = _handlers[pageId];
      // print('messageHandler: $handler');
      if (null != handler) {
        handler.onMessage(jsonObj);
      } else {
        // 实时调试socket过来的数据
        var jsonObject = jsonDecode(jsonObj['message']);
        var pageCode = jsonObject['pageCode'];
        var content = jsonObject['content'];
        _pages.putIfAbsent(pageCode, () => content);
        _handlers.forEach((k, v) {
          if (k.startsWith(pageCode)) {
            v.onMessage(jsonObj);
          }
        });
      }
      return Future<String>.value("success");
    });
  }

  void _socket(Map<String, dynamic> map) {
    if (null == context) return;
    if (null != _factory) {
      _factory.clear();
      _callOnUnload();
    }
    var jsonObject = jsonDecode(map['message']);
    var pageCode = jsonObject['pageCode'];
    var content = jsonObject['content'];
    if (_pageCode == pageCode) {
      _data = jsonDecode(content);
      _factory.clear();
      _create();
    }
  }

  Future _update(Map<String, dynamic> map) async {
    var jsonObject = jsonDecode(map['message']);
    await _factory.updateTree(jsonObject);
  }

  void _updateTitle(Map<String, dynamic> map) {
    setState(() {
      _title = map['message'];
    });
  }

  void _setNavigationBarColor(Map<String, dynamic> map) {
    setState(() {
      _appBarColor = parseColor(map['message'], defaultValue: Colors.blue);
    });
  }

  void _setBackgroundColor(Map<String, dynamic> map) {
    setState(() {
      _backgroundColor =
          parseColor(map['message'], defaultValue: Colors.grey[200]);
    });
  }

  void _navigateTo(Map<String, dynamic> map) {
    var jsonObject = jsonDecode(map['message']);
    var url = jsonObject['url'];
    if (null != url) {
      var uri = Uri.parse(url);
      var path = uri.path;
      var params = uri.queryParameters;
      if (null != path && path.isNotEmpty) {
        var args = Map<String, dynamic>()
          ..putIfAbsent("pageCode", () => path)
          ..putIfAbsent("args", () => params);

        Navigator.of(context).push(MaterialPageRoute(
            builder: (context) => DpoMiniAppPage(args, _pages, _handlers,
                _methodChannel, _basicChannel, _indexPage)));
      }
    }
  }

  void _create() async {
    var body = _data['body'];
    var styles = _data['style'];
    var script = _data['script'];
    var config = _data['config'];
    if (null == script) {
      script = "";
    }

    _initConfig(config);
    _initScript(script);
    _callOnLoad();
    var component = await _factory.createComponentTree(null, body, styles);
    var tree = await _factory.createWidgetTree(null, component);
    setState(() {
      _tree = tree;
    });
  }

  void _initConfig(Map<String, dynamic> config) {
    if (null == config) {
      return;
    }
    _title = config['navigationBarTitleText'];
    _appBarColor = parseColor(config['navigationBarBackgroundColor'],
        defaultValue: Colors.blue);
    _backgroundColor =
        parseColor(config['backgroundColor'], defaultValue: Colors.grey[200]);
    _enablePullDownRefresh = config['enablePullDownRefresh'];
  }

  @override
  void initState() {
    super.initState();
    _data = jsonDecode(_pages[_pageCode]);
    _create();
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
//    print("lifecycle didChangeDependencies $_pageId");
  }

  @override
  void dispose() {
    super.dispose();
//    print("lifecycle dispose $_pageId");
    _handlers.remove(_pageId);
    _factory.clear();
    _callOnUnload();
  }

  void _initScript(String script) {
    _methodChannel.invokeMethod(
        "attachPage", {"pageId": _pageId, "script": decodeBase64(script)});
    _methodChannel.invokeMethod("initComplete", {"pageId": _pageId});
  }

  void _callOnLoad() {
    _methodChannel
        .invokeMethod("onLoad", {"pageId": _pageId, "args": jsonEncode(_args)});
  }

  void _callOnUnload() {
    _methodChannel.invokeMethod("onUnload", {"pageId": _pageId});
  }

  void _startPullDownRefresh(Map<String, dynamic> map) {
    if (null != _refreshIndicatorKey) {
      _refreshIndicatorKey.currentState.show();
    }
  }

  void _stopPullDownRefresh(Map<String, dynamic> map) {
    if (null != _completer) {
      _completer.complete(true);
    }
  }

  Future<void> _onRefresh() {
    _methodChannel.invokeMethod("onPullDownRefresh", {"pageId": _pageId});
    _completer = new Completer<bool>();
    return _completer.future.then((bool success) {
      return success;
    });
  }

  @override
  Widget build(BuildContext context) {
    var child = _tree;
    if (null != _enablePullDownRefresh &&
        _enablePullDownRefresh &&
        null != _tree) {
      _refreshIndicatorKey = GlobalKey<RefreshIndicatorState>();
      child = RefreshIndicator(
          key: _refreshIndicatorKey, onRefresh: _onRefresh, child: _tree);
    }

    return Scaffold(
        key: PageStorageKey(_tree),
        appBar: AppBar(
          title: Text(_title),
          centerTitle: true,
          backgroundColor: _appBarColor,
        ),
        backgroundColor: _backgroundColor,
        body: child);
  }

  @override
  void onMessage(Map<String, dynamic> message) {
    int type = message['type'];
    switch (type) {
      case 0: //socket
        _socket(message);
        break;
      case 1: //onclick setData
        _update(message);
        break;
      case 2: //set_navigation_bar_title
        _updateTitle(message);
        break;
      case 3: //navigate_to
        _navigateTo(message);
        break;
      case 4: //set_navigation_bar_color
        _setNavigationBarColor(message);
        break;
      case 5: //set_background_color
        _setBackgroundColor(message);
        break;
      case 6: //start_pull_down_refresh
        _startPullDownRefresh(message);
        break;
      case 7: //stop_pull_down_refresh
        _stopPullDownRefresh(message);
        break;
    }
  }
}
